// MMSQ: 2048 30 1 256

__kernel void
multi_square_2048_30_1(
	__global uint *gx,
	__constant uint *m,
	uint m0i, int w)
{
	uint ga;
	uint x[69], d[69];
	uint m0il, m0ih;

	/*
	 * Read value from global memory.
	 */
	ga = get_global_id(0) * 69;
	#pragma unroll 69
	for (int i = 0; i < 69; i ++) {
		x[i] = gx[ga + i];
	}
	m0il = m0i & 0x7FFF;
	m0ih = (m0i >> 15) & 0x7FFF;

	#pragma unroll 1
	for (int k = 0; k < w; k ++) {
		ulong dh;
		uint cc;
		bool nsub;
		uint x0, x0l, x0h;

		for (int i = 0; i < 69; i ++) {
			d[i] = 0;
		}
		x0 = x[0];
		x0l = x0 & 0x7FFF;
		x0h = x0 >> 15;
		dh = 0;
		#pragma unroll 1
		for (int i = 0; i < 69; i ++) {
			uint xi, xil, xih;
			uint f, fl, fh;
			uint rl, rh;
			uint d0, d1;
			uint tmm;

			xi = x[i];
			xil = xi & 0x7FFF;
			xih = xi >> 15;

			f = mad24(xil, x0l, d[0])
				+ (mad24(xil, x0h, mul24(xih, x0l)) << 15);
			fl = f & 0x7FFF;
			fh = (f >> 15) & 0x7FFF;
			f = mad24(fl, m0il,
				(mad24(fl, m0ih, mul24(fh, m0il)) << 15));
			fl = f & 0x7FFF;
			fh = (f >> 15) & 0x7FFF;

			d0 = d[0] & 0x7FFF;
			d1 = d[0] >> 15;
			#pragma unroll 69
			for (int j = 0; j < 69; j ++) {
				uint xj, mj;
				uint xjl, xjh, mjl, mjh;

				xj = x[j];
				xjl = xj & 0x7FFF;
				xjh = xj >> 15;
				mj = m[j];
				mjl = mj & 0x7FFF;
				mjh = mj >> 15;

				rl = mad24(fl, mjl, mad24(xil, xjl, d0) + rl);
				d0 = rl & 0x7FFF;
				rl >>= 15;
				rl = mad24(fl, mjh, mad24(xil, xjh, d1) + rl);
				d1 = rl & 0x7FFF;
				rl >>= 15;

				rh = mad24(fh, mjl, mad24(xih, xjl, d1) + rh);
				d1 = rh & 0x7FFF;
				rh >>= 15;
				if (j != 0) {
					d[j - 1] = d0 | (d1 << 15);
				}
				if (j == 68) {
					d0 = dh;
					d1 = 0;
				} else {
					uint dj;

					dj = d[j + 1];
					d0 = dj & 0x7FFF;
					d1 = dj >> 15;
				}
				rh = mad24(fh, mjh, mad24(xih, xjh, d0) + rh);
				d0 = rh & 0x7FFF;
				rh >>= 15;
			}
			rl += d0;
			d0 = rl & 0x7FFF;
			rl >>= 15;
			rl += d1;
			d1 = rl & 0x7FFF;
			rl >>= 15;
			rh += d1;
			d1 = rh & 0x7FFF;
			rh >>= 15;
			d[68] = d0 | (d1 << 15);
			dh = rl + rh;
		}

		/*
		 * Copy and possibly subtract the modulus to ensure that
		 * the value fits.
		 */
		nsub = dh != 0;
		cc = 0;
		#pragma unroll 69
		for (int i = 0; i < 69; i ++) {
			uint xi;

			xi = d[i];
			if (nsub) {
				uint yi;

				yi = xi - m[i] - cc;
				cc = (yi >> 30) & 1;
				xi = yi & 0x3FFFFFFF;
			}
			x[i] = xi;
		}
	}

	/*
	 * Write back result into global memory.
	 */
	#pragma unroll 69
	for (int i = 0; i < 69; i ++) {
		gx[ga + i] = x[i];
	}
}
