// MMSQ: 2048 31 1 256

__kernel void
multi_square_2048_31_1(
	__global uint *gx,
	__constant uint *m,
	uint m0i, int w)
{
	uint ga;
	uint x[67], d[67];

	/*
	 * Read value from global memory.
	 */
	ga = get_global_id(0) * 67;
	#pragma unroll 67
	for (int i = 0; i < 67; i ++) {
		x[i] = gx[ga + i];
	}
	m0i &= 0x7FFFFFFF;

	#pragma unroll 1
	for (int k = 0; k < w; k ++) {
		ulong dh;
		uint cc, r3;
		bool nsub;

		for (int i = 0; i < 67; i ++) {
			d[i] = 0;
		}
		dh = 0;
		#pragma unroll 1
		for (int i = 0; i < 67; i ++) {
			uint f, xi;
			ulong r, zh;

			xi = x[i];
			f = (d[0] + xi * x[0]) * m0i;
			f &= 0x7FFFFFFF;
			r = 0;
			#pragma unroll 67
			for (int j = 0; j < 67; j ++) {
				ulong z;

				z = (ulong)d[j] + (ulong)xi * (ulong)x[j]
					+ (ulong)f * (ulong)m[j] + r;
				r = z >> 31;
				if (j != 0) {
					d[j - 1] = (uint)z & 0x7FFFFFFF;
				}
			}
			zh = dh + r;
			d[66] = (uint)zh & 0x7FFFFFFF;
			dh = zh >> 31;
		}

		/*
		 * Copy and possibly subtract the modulus to ensure that
		 * the value fits.
		 */
		nsub = dh != 0;
		cc = 0;
		#pragma unroll 67
		for (int i = 0; i < 67; i ++) {
			uint xi;

			xi = d[i];
			if (nsub) {
				uint yi;

				yi = xi - m[i] - cc;
				cc = yi >> 31;
				xi = yi & 0x7FFFFFFF;
			}
			x[i] = xi;
		}
	}

	/*
	 * Write back result into global memory.
	 */
	#pragma unroll 67
	for (int i = 0; i < 67; i ++) {
		gx[ga + i] = x[i];
	}
}
