// MMSQ: 4096 15 8 256

__kernel __attribute__((reqd_work_group_size(256,1,1))) void
multi_square_4096_15_8(
	__global uint *gx,
	__constant uint *gm,
	uint m0i, int w)
{
	bool low_chunk, high_chunk;
	uint ga, la, lo, lma;
	uint x[35];
	uint d[35];
	uint m[35];
	__local uint cxi[256], cf[256], cr[256], cu[256];

	/*
	 * Read values from global memory.
	 */
	la = get_local_id(0);
	lma = la & 7;
	low_chunk = (lma == 0);
	high_chunk = (lma == 7);
	lo = 35 * lma;
	ga = (get_global_id(0) >> 3) * 280 + lo;
	#pragma unroll 35
	for (int i = 0; i < 35; i ++) {
		x[i] = gx[ga + i];
		m[i] = gm[lo + i];
	}

	m0i &= 0x7FFF;
	for (int k = 0; k < w; k ++) {
		uint dh;
		bool nsub;

		/*
		 * Clear the destination buffer.
		 */
		#pragma unroll 35
		for (int i = 0; i < 35; i ++) {
			d[i] = 0;
		}
		cr[la] = 0;
		cu[la] = 0;
		barrier(CLK_LOCAL_MEM_FENCE);

		dh = 0;
		#pragma unroll 1
		for (int i = 0; i < 287; i ++) {
			int id;
			uint xi;
			uint f;
			bool do_run;

			/*
			 * Each chunk must send its xi and f values to the
			 * next chunk. The new xi (for the low chunk) is
			 * computed by the relevant thread.
			 */
			xi = cxi[la];
			f = cf[la];
			barrier(CLK_LOCAL_MEM_FENCE);
			if (!high_chunk) {
				cxi[la + 1] = xi;
				cf[la + 1] = f;
			}
			id = (i * 1873) >> 16;
			if (lma == id) {
				cxi[la - lma] = x[i - lo];
			}
			barrier(CLK_LOCAL_MEM_FENCE);
			xi = cxi[la];
			if (low_chunk) {
				f = mul24(mad24(xi, x[0], d[0]) & 0x7FFF, m0i);
				f &= 0x7FFF;
				cf[la] = f;
			} else {
				f = cf[la];
			}

			/*
			 * All chunks run except at the start and at the end.
			 */
			do_run = ((i >= lma) && (i < (280 + lma)));
			if (do_run) {
				uint r;

				r = cr[la];
				#pragma unroll 18
				for (int j = 0; j < 18; j ++) {
					r = mad24(xi, x[j], d[j]) + r;
					r = mad24(f, m[j], r);
					if (j == 0) {
						if (!low_chunk) {
							cu[la - 1] = r & 0x7FFF;
						}
					} else {
						d[j - 1] = r & 0x7FFF;
					}
					r >>= 15;
				}
				barrier(CLK_LOCAL_MEM_FENCE);
				if (!high_chunk) {
					d[34] = cu[la];
				}
				#pragma unroll 17
				for (int j = 18; j < 35; j ++) {
					r = mad24(xi, x[j], d[j]) + r;
					r = mad24(f, m[j], r);
					d[j - 1] = r & 0x7FFF;
					r >>= 15;
				}
				if (high_chunk) {
					uint z;

					z = dh + r;
					d[34] = z & 0x7FFF;
					dh = z >> 15;
				} else {
					cr[la + 1] = r;
				}
			}
			barrier(CLK_LOCAL_MEM_FENCE);
		}

		/*
		 * Propagate last word for all chunks (except the high one,
		 * which is special).
		 */
		if (!high_chunk) {
			d[34] = cu[la];
		}

		/*
		 * High chunk decides whether there is a subtraction to
		 * perform. The "do it" flag is stored in the cf[] array.
		 */
		if (high_chunk) {
			uint ts;

			ts = dh != 0;
			cf[la] = ts;
			cf[la - 1] = ts;
			cf[la - 2] = ts;
			cf[la - 3] = ts;
			cf[la - 4] = ts;
			cf[la - 5] = ts;
			cf[la - 6] = ts;
			cf[la - 7] = ts;
		}
		cr[la] = 0;
		barrier(CLK_LOCAL_MEM_FENCE);

		/*
		 * Copy and possibly subtract the modulus to ensure that
		 * the value fits. This must be done in several loops
		 * because each chunk needs the carry from the previous
		 * one.
		 */
		nsub = cf[la];
		#pragma unroll 1
		for (int i = 0; i < 8; i ++) {
			if (i == lma) {
				uint cc;

				cc = cr[la];
				barrier(CLK_LOCAL_MEM_FENCE);
				#pragma unroll 35
				for (int j = 0; j < 35; j ++) {
					uint xj;

					xj = d[j];
					if (nsub) {
						uint yj;

						yj = xj - m[j] - cc;
						cc = (yj >> 15) & 1;
						xj = yj & 0x7FFF;
					}
					x[j] = xj;
				}
				if (!high_chunk) {
					cr[la + 1] = cc;
					barrier(CLK_LOCAL_MEM_FENCE);
				}
			}
		}
	}

	/*
	 * Write back result into global memory.
	 */
	#pragma unroll 35
	for (int i = 0; i < 35; i ++) {
		gx[ga + i] = x[i];
	}
}
